#!/bin/bash
# SPDX-License-Identifier: MIT

cd "$(dirname "$0")" || exit

error() {
	[ -n "$timespec" ] && echo -n "$base / $timespec: " >&2
	printf "%s\n" "$@" >&2
	exit 1
}

build_faketime() {
	[ -e faketime/libfaketime.so.1 ] && return
	(
		cd faketime || exit
		cc -o libfaketime.so.1 -Wl,-soname,libfaketime.so.1  \
			-Wl,--version-script=libfaketime.map -shared \
			-std=gnu99 -Wall -Wextra -Werror -DFAKE_PTHREAD \
			-DFAKE_STAT -DFAKE_UTIME -DFAKE_SLEEP -DFAKE_TIMERS \
			-DFAKE_INTERNAL_CALLS -fPIC -Wno-nonnull-compare \
			libfaketime.c -ldl -lm -lrt -lpthread
	) || exit
}

run_date() {
	# date -d can set an offset, but we don't want to mess with tests:
	# use libfaketime
	LD_PRELOAD=./faketime/libfaketime.so.1 \
		FAKETIME="$base" \
		date -d "$timespec" +%s
}

run_schedts() {
	../../tools/schedule_ts -B "$base" "$timespec"
}

compare_outputs() {
	local timespec="$1"
	local date schedts

	date="$(run_date 2>/dev/null)" || error "date failed"
	schedts="$(run_schedts)" || error "schedule_ts failed"

	[ "$date" = "$schedts" ] \
		|| error "Bad timestamp $schedts ($(date -d @"$schedts")) instead of $date ($(date -d @"$date"))"
}

compare_realtime() {
	# check time obtention works (without -B)
	local timespec="$1"

	local date schedts
	date=$(date -d "$timespec" +%s)
	schedts=$(../../tools/schedule_ts "$timespec")

	# hopefully won't take more than 10 seconds to run even with slow emulation...
	[ "$((date > schedts ? date - schedts : schedts - date))" -lt 10 ] \
		|| error "Bad timestamp $schedts ($(date -d @"$schedts")) instead of $date ($(date -d @"$date"))"
}

compare_realtime now
compare_realtime tomorrow

build_faketime

RANDOM_SEED=${RANDOM_SEED:-$RANDOM}
RANDOM=$RANDOM_SEED
echo "Running with RANDOM_SEED=$RANDOM_SEED"

for _ in {1..10}; do
	base="$(date -d "@$((RANDOM*65535+RANDOM))" +"%Y-%m-%d %H:%M:%S")"
	echo "base $(date -d "$base")"
	# basic tests
	compare_outputs now
	compare_outputs "next week"
	compare_outputs "0 tomorrow"
	compare_outputs "3 2 days"
	compare_outputs "5 pm next week"
	compare_outputs "705 2 weeks"
	compare_outputs "8:12 next month"
	compare_outputs "09:04:23 pm 2 months"
	compare_outputs "23 next monday"
	compare_outputs "4am 2 monday"
	compare_outputs "2 min 3 secs"
	compare_outputs "now + 3 hour"
	compare_outputs "3 hour 2 year"
	compare_outputs "oct 3 3 2012"
	compare_outputs "3 oct 2012 1712"

	# all check days of the week
	for day in sunday mon tue wed thu friday sat; do
		compare_outputs "$day 3 3 min ago"
		compare_outputs "next $day"
		compare_outputs "3 $day 1"
		compare_outputs "3 last $day"
	done
done

# check all months just once
for month in jan february march apr may june july aug sept oct nov dec; do
	compare_outputs "01 $month"
done

# TODO some random stuff?
